// The following program was written as a part of a Master's thesis in computational physics at NTNU spring 2016.
// Special thanks goes to Morten Vassvik, which was very helpful in the making of this program, with the visualizaion in particular.

// The program takes three arguments: Lx=lattice size, cutoff,samples.
// GLFW visualization may be enabled uncommenting #define GRAPHICS further down
// GLFW is an OpenGL library that may be downloaded from www.glfw.org 

// The seed is stored in binary format in the file "seed_binary.bin". If such a file already exist, the program first reads the seed from this file
// It then skips the amount of numbers that the current simulation will need, and stores that value at the end of the same file.

// Compilation:
//      For compilation without graphics on unix:
//      g++ -O3  -o lls lls.cpp -std=c++11

//      For compilation with graphics on Mac OS X:
//      g++ -O3 -pg -o lls lls.cpp -framework Cocoa -framework OpenGL -framework IOKit -I/usr/local/include/ -lglfw3 -framework Corevideo -std=c++11 -framework GLUT -g -fprofile-arcs -ftest-coverage 

// Example of useage:
//      To run the program with Lx=128, cutoff=0.7, and samples=50:
//      ./lls 128 0.7 50

// The most interesting function in this program is break(), which breaks the selected fiber, and updates the system

// Note that each fiber is a node in the AVL-tree for up to four perimeters
// Each fiber also have a pointer to a node in the tree for fibers in many perimeters, that is NULL by default
// Moreover, each perimeter is also a node in the tree for the perimeters
// And each perimeter also points to the head of a linked list of fibers that are included in at least one additional perimeter
// The implementation of the AVL tress and the linked list for fibers in multiple perimeters is moved to the separate file lls_data_structures.h to increase readability.

// Further development:
//      The AVL structure ought preferably to be implemented with templates and overloaded operators.
//      However, std::set might be a better alternative altogheter.
//      The fibers in many perimeters constitute the bottle neck of the program. 
//      The time step at which the data structures for the fibers in many periemters is swapped is not yet completely optimized.
//      To save data, uncomment save_to_file. 
//      This function stores the strain curve, the burst distribution, the cluster count and the fiber elongation to file.
//      Storing the cluster count as a histogram (similar to the data of the next weakest fiber) would reduce file size considerably.

// During visualization mode: 
//      Esc closes the window
//      Space break 1% of the fibers (LLS). 
//      r breaks all the fibers (LLS). 
//      o breaks the fiber with the lowest failure threshold (ELS).
//      tab breaks all the fibers (ELS)
//      2 swaps to visualization of the AVL trees. Swap back with 1.
//      right mouse click breaks the fiber.
//      left mouse click is used for naviation
//      Mouse scrolling is used for zooming in and out
//
//
// Eivind Bering
// eivind.bering@gmail.com

#include <stdio.h>
#include <math.h>
#include <vector> 
#include <algorithm>
#include <cstdlib>
#include "xorshift.h"    //This file is needed for compilation, and is found together with the current file.
#include <map>
#include <unistd.h>
#define GRAPHICS   //Uncomment this line to compile with graphics

double cutoff = 0.617;    //cutoff in the exponential proabilitiy distribution with cutoff P(x)=1-exp(cutoff-x)
int Lx = 256;             //system size
int Ly=Lx;
int N=Lx*Ly;
int samples=100;
double treeToVec;   // The timestep k/N at which the fibers in many perimeters is stored in a vector instead of an AVL tree
int clusterCount=0;
int maxClusterCount=0;
int nextWeakestFiber=0;
bool nwf_broken=false;  //Is the next weakest fiber broken?
clock_t tot_time = clock();
std::vector<double> straincurve;
std::vector<double> fiberElongation;
char strainfilename [100];
char ccfilename [100];  
char burstfilename [100];
char fburstfilename [100];
char nwffilename [100];
char stressfilename [100];
enum FibreStatus {INTACT = 1, BROKEN = 0};
double time_since_start = 0;
double time_at_start    = 0;
struct Fiber;
struct ManyPerimeterLinkedList;
struct ManyPerimeter;

struct Perimeter{

    int brokenFibers;
	int intactFibers;
	Fiber *root;       //The root fiber in this particular perimeter tree
    
    //Each perimeter is a node in the tree of perimeters.
    unsigned char height;   
    Perimeter* left;
    Perimeter* right;

    ManyPerimeterLinkedList *manyPerHead;     //The fibers in this perimeter that are in more than one perimeter. Pointer to the head of that list.

    Perimeter(){        
        manyPerHead=NULL; 
    	left=NULL;
    	right=NULL;
    	height=1;
        root=NULL;
    };

	Fiber *weakest(); //The fiber with the hightest ratio of load to threshold
	double weakestLoad();
	double loadshare(){return brokenFibers/(double)intactFibers;};
};

struct Fiber{
	Perimeter *myPerimeter[4];
	double t;  //The INVERSE fiber threshold t=1/x

    double trueload();

	int status;
	int fiberID;
	
    //Each fiber is also a node in up to four perimeter trees
	Fiber *left[4]; 
	Fiber *right[4]; 
	unsigned char height[4];

    ManyPerimeter *manyPerimeterNode;
	
    Fiber() {
        status=INTACT;
        manyPerimeterNode=NULL;
    };

	Fiber* getRight(Perimeter *p);
	Fiber* getLeft(Perimeter *p);
	void setRight(Fiber *f, Perimeter *p);
	void setLeft(Fiber *f, Perimeter *p);
	void setNewPer(Perimeter *myPer);
	void removeFromPer(Perimeter *myPer);
	int perimeterCount();
	bool inPerimeter(Perimeter *myPer);
	void Break();
	void replacePer(Perimeter *otherPer,Perimeter *myPer);//Replace perimeter of current fiber
};

struct ManyPerimeterLinkedList          //Each perimeter has a linked list connecting the fibers in many perimeters
{
    Fiber *f;
    ManyPerimeterLinkedList *next;
    ManyPerimeterLinkedList(){
        next=NULL;
    };
};

struct ManyPerimeter
{
    Fiber *f;
    double trueload;
    unsigned char height;
    ManyPerimeter* left;
    ManyPerimeter* right;

    ManyPerimeter(){
        height=1;
        left=NULL;
        right=NULL;
    };
};

ManyPerimeter *manyPerimeterRoot=NULL;
 
std::vector<Fiber *> noperimeter;       //All (intact) fibers not in noPerLoad perimeter. Sorted by threshold.
std::vector<Fiber *> manyPerimeters;    //vector of fibers in many perimeters

Perimeter *rootPerimeter=NULL;
Fiber *fibers;
Fiber *weakestFiberManyPer=NULL;   //The fiber in many perimeters with the highest ratio of load to fiber threshold 
double weakestFiberManyPerLoad=0;   
int totBreaks = 0;
bool lowestThreshold(Fiber *f1,Fiber *f2);

void breakWeakest();    //Breaks the fiber with the highest load to threshold ratio (LLS)
void makeNewSeedFile(); //If the file "seed_binary.bin" don't exist, it will be created
void initializeSeed();  //Read current seed, and write seed for next simulation
void save_to_file(int sample,std::vector<double> &meanStraincurve,std::vector<double> &stdstrain,std::map<int, int> &burst,std::vector<int> &new_cc,std::map<int, int> &hist,std::map<int, int> &fatalBurst,std::vector<double> &meanFiberElongation);

double Fiber::trueload(){
    double ls=0;   //loadshare
    for (int i = 0; i < 4; ++i)
    {
        if(myPerimeter[i] != NULL){
            ls+=myPerimeter[i]->loadshare();
        }
    }
    return t*(1+ls);    
}
//////////////////////////////////////////////////////////////////////////////////////////////////////////
#include "lls_data_structures.h" //Implementation of AVL-trees for the fibers, the perimeters, 
// the fibers in many perimeters and the linked list of fibers in many perimeters
//////////////////////////////////////////////////////////////////////////////////////////////////////////

bool lowestThreshold(Fiber *f1,Fiber *f2){
	return (f1->t) < (f2->t);
}
bool lowestLoad(Fiber *f1,Fiber *f2){
	double f1_loadshare=0;
	double f2_loadshare=0;
	for(int i=0;i<4;i++){
		if(f1->myPerimeter[i]!=NULL){
			f1_loadshare+=f1->myPerimeter[i]->loadshare();
		}
		if(f2->myPerimeter[i]!=NULL){
			f2_loadshare+=f2->myPerimeter[i]->loadshare();
		}
	}
	
	return ( ( (f1_loadshare+1)*f1->t  ) < (f2_loadshare+1)*f2->t );
}
int Fiber::perimeterCount(){
	int count=0;
	for (int i = 0; i < 4; ++i)
	{
		if(myPerimeter[i]!=NULL){
			count+=1;
		}
	}
	return count;
}
void Fiber::setNewPer(Perimeter *myPer){
	for (int i=0;i<4;i++){
		if(myPerimeter[i]==myPer) {return;}
		else if (myPerimeter[i]==NULL){
			myPerimeter[i]=myPer;
			return;
		}
	}
}
void Fiber::removeFromPer(Perimeter *otherPer){//Removes fiber from other perimeter
	for (int i=0;i<4;i++){
		if(myPerimeter[i]==otherPer) {
			myPerimeter[i]=NULL;
			left[i]=NULL;
			right[i]=NULL;
			height[i]=1;
		}
	}
}
void Fiber::replacePer(Perimeter *otherPer,Perimeter *myPer){//Replace perimeter of current fiber
		
	for(int i=0;i<4;i++){
		if(myPerimeter[i]==otherPer) {
			myPerimeter[i]=myPer;
			left[i]=NULL;
			right[i]=NULL;
			height[i]=1;
		}
	}
}
bool Fiber::inPerimeter(Perimeter *myPer){
	for (int i = 0; i < 4; ++i)
	{
		if(myPerimeter[i]==myPer) return true;
	}
	return false;
}
//Merge perimeter oldPer with root fiber n and perimeter p with root fiber f
void postOrderMerge(Fiber* n, Perimeter* p,Perimeter* oldPer){
    if (n == NULL) return;
    postOrderMerge(n->getLeft(oldPer),p,oldPer);
    postOrderMerge(n->getRight(oldPer),p,oldPer);
 	if (!(n->inPerimeter(p)) && n->status==INTACT){
        ManyPerimeter *mp;
        n->replacePer(oldPer,p);
    	p->root=insert(p->root,n,p); //insert fiber n in perimeter p 
    	p->intactFibers+=1;
        int pc=n->perimeterCount();
        if(pc>1){
            if(totBreaks<int(N*treeToVec)){ //if neighbours is in many perimeters, remove from manyperimetertree, and insert in the linked list(s)
                mp=n->manyPerimeterNode;
    
                if(mp){ 
                    manyPerimeterRoot=remove(manyPerimeterRoot,mp);
                    mp->height=1;
                    mp->left=NULL;
                    mp->right=NULL;
                }
                insert(n,p);
            }else {
                manyPerimeters.push_back(n);
            }
        }
    }else{
        n->removeFromPer(oldPer);
    }
}
void treeToVector(ManyPerimeter *mp){  //Move the fibers in manyperimeters from the AVL tree to the vector
    if(!mp) return;
    treeToVector(mp->left);
    treeToVector(mp->right);
    if(mp->f->status==INTACT) manyPerimeters.push_back(mp->f);
    mp->f->manyPerimeterNode=NULL;
    delete mp;
}
void deleteAllLists(Perimeter *p){
    if(!p)return;
    deleteAllLists(p->left);
    deleteAllLists(p->right);
    
    deleteList(p);
    p->manyPerHead=NULL;
}
void updateManyPerVector(){
    weakestFiberManyPer  =NULL;
    weakestFiberManyPerLoad =0;
    double thisLoad=0;

    std::vector< Fiber * >::iterator it;
    it = manyPerimeters.begin();

    while(it!=manyPerimeters.end()) {
       
        if((*it)->status==BROKEN || (*it)->perimeterCount() < 2){
            manyPerimeters.erase(it);

        }else{
            thisLoad=(*it)->trueload();
            if(thisLoad > weakestFiberManyPerLoad){
                weakestFiberManyPer=(*it);
                weakestFiberManyPerLoad=thisLoad;
            }
            ++it;   
        }
    }
}
void Fiber::Break(){  //breaks the fiber, and updates the system to make it easy for breakWeakest() to find the next fiber to break
	if (status==INTACT){
		status=BROKEN;
	}else{
		return;
	}
	totBreaks+=1;
    if(totBreaks==N){
        Perimeter *oldRoot=rootPerimeter;
        rootPerimeter=remove(rootPerimeter,rootPerimeter);
        delete oldRoot;
        manyPerimeters.clear();
        return;
    } 
    if(totBreaks==int(N*treeToVec)){ //switch from trees to vector
        printf("\r ************   Switching from AVL to vector for the fibers in many perimeters!   ************ \n\n");
        treeToVector(manyPerimeterRoot);
        deleteAllLists(rootPerimeter);
        manyPerimeterRoot=NULL;
    }
	int xx = fiberID % Lx;
	int yy = fiberID / Lx;
    Fiber *neighbours[4];
	Perimeter *myPer=NULL; //current perimeter
    ManyPerimeter *mp;
    ManyPerimeterLinkedList *m;
    int pc;                 //perimeter count
	int intactNeighbours=0;
    //Find the neighbours
	neighbours[0]=&fibers[((yy + 0 + Ly) % Ly)*Lx + ((xx + 1 + Lx) % Lx)];
	neighbours[1]=&fibers[((yy + 1 + Ly) % Ly)*Lx + ((xx + 0 + Lx) % Lx)];
	neighbours[2]=&fibers[((yy + 0 + Ly) % Ly)*Lx + ((xx - 1 + Lx) % Lx)];
	neighbours[3]=&fibers[((yy - 1 + Ly) % Ly)*Lx + ((xx + 0 + Lx) % Lx)];

	//Count intact neighbours
	for (int i=0;i<4;i++){
		if (neighbours[i]->status==INTACT){
			intactNeighbours+=1;
		}
	}
	if (intactNeighbours==4){
		myPer = new Perimeter;
		myPer->brokenFibers=1;
		myPer->intactFibers=4;
		myPer->root    = NULL;
        clusterCount++;
        if(clusterCount>maxClusterCount) maxClusterCount=clusterCount;
		
        if(totBreaks<int(N*treeToVec)){ //update trees if less than treeToVec of the fibers have broken
            for(int i=0;i<4;i++){
		  	    pc=neighbours[i]->perimeterCount();
                if(pc>1){ //if neighbours is in many perimeters, remove from manyperimetertree, and insert in the linked list(s)
                    mp=neighbours[i]->manyPerimeterNode;
                    if(!mp){
                        printf("one of 4 intact neighbour with perimetercount >1 has no manyPerimeterNode! ID %d \n",neighbours[i]->fiberID);
                        exit(-1);
                    }
                    manyPerimeterRoot=remove(manyPerimeterRoot,mp);
                    mp->height=1;
                    mp->left=NULL;
                    mp->right=NULL;
                }
                
                if(pc==1){
                    for (int j= 0; j < 4; ++j)
                    {
                        if(neighbours[i]->myPerimeter[j]!=NULL){
                            insert(neighbours[i],neighbours[i]->myPerimeter[j]);
                        }
                    }
                }
                neighbours[i]->setNewPer(myPer);
                if(pc>0){
                    insert(neighbours[i],myPer); //insert in the linked list of many periemters
                }
                myPer->root=insert(myPer->root,neighbours[i],myPer);
                     //insert in the linked list of many perimeters
		  }
		} else{ // update the vector 
            for(int i=0;i<4;i++){
                neighbours[i]->setNewPer(myPer);
                myPer->root=insert(myPer->root,neighbours[i],myPer);

                if(neighbours[i]->perimeterCount()>1){
                    manyPerimeters.push_back(neighbours[i]);
                }
            }
        }
		rootPerimeter=insert(rootPerimeter,myPer); //insert the intact neighbours in the tree of perimeter myPer
	} else{
		Fiber *f;
		Perimeter *otherPer;
		Perimeter *p;
		double weakestStrength=0;
		int largestPer=0;
		bool inBrokenPerimeter;
        ManyPerimeterLinkedList *n;

		for(int k=0;k<4;k++){ //Remove all broken perimeters from the perimeter tree
			if(myPerimeter[k]!=NULL){
				otherPer=myPerimeter[k];
				rootPerimeter=remove(rootPerimeter,otherPer);
				otherPer->height=1;
				otherPer->left=NULL;
				otherPer->right=NULL;
			}
		}

		//Find largest broken perimeter, and name it myPer
		for (int i=0;i<4;i++){
			if (myPerimeter[i]!=NULL){
				if(myPerimeter[i]->intactFibers > largestPer){
					myPer=myPerimeter[i];
					largestPer=myPer->intactFibers;
				}
			}
		}
		myPer->root=remove(myPer->root,this,myPer);

        if(totBreaks<int(N*treeToVec)){ //remove the affected fibers from the many perimeter tree
            m=myPer->manyPerHead;
            while (m){
                n=m->next;
                f=m->f;
                mp=f->manyPerimeterNode;
                if(mp) {
                    manyPerimeterRoot = remove(manyPerimeterRoot,mp);
                    mp->height=1;
                    mp->left=NULL;
                    mp->right=NULL;
                } //remove from manyperimeter tree
                m=n;
            } 
        }
		myPer->brokenFibers+=1;
		myPer->intactFibers-=1;
		//check the neighbours, if they are not included in a broken perimeter, they are inserted in perimeter myPer
		for (int k=0;k<4;k++){
			f=neighbours[k];
			if(f->status==INTACT){
				inBrokenPerimeter=false;
				for(int j=0;j<4;j++){
						if(f->myPerimeter[j]!=NULL){
							for(int i=0;i<4;	i++){
								if (f->myPerimeter[j]==myPerimeter[i]){
									inBrokenPerimeter=true;
								}
							}
						}
				}
				if (not inBrokenPerimeter){

                    pc=f->perimeterCount();

                    if(totBreaks<int(N*treeToVec)){//update manyperimeter tree
                        if(pc>1){ //if neighbours is in many perimeters, remove from manyperimetertree, and insert in the linked list(s)
                            mp=f->manyPerimeterNode;
                            manyPerimeterRoot=remove(manyPerimeterRoot,mp);
                            mp->height=1;
                            mp->left=NULL;
                            mp->right=NULL;
                        }
    
                        if(pc==1){
                            //insert in the manyperimeterlinkedlist of myPer and the other Perimeter
                            for (int i = 0; i < 4; ++i)
                            {
                                if(f->myPerimeter[i]!=NULL){
                                    insert(f,f->myPerimeter[i]);
                                }
                            }
                        }
                        f->setNewPer(myPer);
                        if (pc>0) insert(f,myPer);
					}else{//update manyperimeter vector 
                        if(pc>0){
                            manyPerimeters.push_back(f);
                        }
                        f->setNewPer(myPer);
                    }
                    myPer->root=insert(myPer->root,f,myPer);
					myPer->intactFibers+=1;
				}

			}
		}
		//Loop over the broken perimeters, insert the fibers in myPer and delete the other broken perimeters
		for(int k=0;k<4;k++){
			if(myPerimeter[k]!=NULL and myPerimeter[k]!=myPer){
				otherPer=myPerimeter[k];	
				postOrderMerge(otherPer->root,myPer,otherPer);
                myPer->brokenFibers+=otherPer->brokenFibers;
                clusterCount--;
                delete otherPer;
			}
		}
		if(myPer->intactFibers > 0){ //update the perimeter tree
			rootPerimeter=insert(rootPerimeter,myPer);
	    }
	}    	
    //The broken fiber is cleared
    for (int i=0;i<4;i++){
        myPerimeter[i]=NULL;
        left[i]=NULL;
        right[i]=NULL;
        height[i]=1;
    }
    if(totBreaks<int(N*treeToVec))    //loop over linked list manyperimeterlinkedlist, remove old lists
    {
        for (int i = 0; i < 4; ++i)
        {
            if(myPerimeter[i]!=NULL and myPerimeter[i]!=myPer){
                deleteList(myPerimeter[i]);
            }
        }
        m=myPer->manyPerHead;
        ManyPerimeterLinkedList *n;
        Fiber *f;
        ManyPerimeterLinkedList *previous=NULL;

        int c=0;
        while(m!=NULL){
            n=m->next;
            f=m->f;
            c+=f->perimeterCount();
            mp=f->manyPerimeterNode;
            if(f==this){
                remove(m,previous,myPer);
                f->manyPerimeterNode=NULL;
                delete mp;
                
            }else if(f->perimeterCount() < 2) {
                remove(m,previous,myPer);
                if(mp){
                    f->manyPerimeterNode=NULL;
                    delete mp;
                }

            }else{
                previous=m;
                if(!mp) {
                    mp= new ManyPerimeter;
                    mp->f=f;
                    mp->trueload=f->trueload();
                    f->manyPerimeterNode=mp;
                    manyPerimeterRoot=insert(manyPerimeterRoot,mp);
                }else{
                    mp->trueload=f->trueload();
                    manyPerimeterRoot=insert(manyPerimeterRoot,mp);
                }
            }
            m=n;
        }
    }
    return;
}

#ifdef GRAPHICS
#include "lls_graphics.h" //GLFW visualization
#else
void runSimulationNoGraphics();
#endif

int main(int argc, char *argv[]) {
    //Read user input
    if(argc>1) {
        Lx=std::atoi(argv[1]);
        Ly=Lx;
        N=Lx*Ly;
    }
    if(argc>2) {
        cutoff=std::atof(argv[2]);
    }
    if(argc>3){
        samples=std::atoi(argv[3]);
    }
    treeToVec=0.43-cutoff*cutoff+0.05*cutoff;  //Timestep at which fibers in manyperimeters in stored in a vector instead of an AVL-tree.
    if(treeToVec<0) treeToVec=0;
    if(argc>4){
        treeToVec=std::atof(argv[4]);
    }

    initializeSeed();//Initialize seed of the xor1024* random number generator with the last seed from file seed_binary.bin
    printf("System size %d \n",N);
    printf("Cutoff is x0=%f, switching many perimeters at k/N = %f \n\n", cutoff,treeToVec);
    double meanNextWeakestFiber=0;
    double oldMeanNextWeakestFiber=0;
    double stdNextWeakestFiber=0;

    //Containers for simulation data, initialized to zero if necessary.
    std::vector<double> meanStraincurve(N,0.0);
    std::vector<double> oldMeanStraincurve(N,0.0);
    std::vector<double> stdstrain(N,0.0);
    std::map<int,int> burst;
    std::vector<double> stdBurst(N,0.0);
    straincurve.resize(N);
    fiberElongation.resize(N);
    std::vector<double> meanFiberElongation(N,0.0);
    std::map<int,int> fatalBurst;
    std::vector<int> maxCC;    //max cluster count 
    std::map<int,int> hist;    //Histogram of how many times the next weakest fiber break at each timestep
    
    #ifdef GRAPHICS
        runWithGraphics();
	#else
        clock_t tot_begin=clock();
        double ct; //current time
        double last_saving_time=0;

        for (int i = 1; i <= samples; ++i)
        {
            ct=(clock()-tot_begin)/(double)CLOCKS_PER_SEC;
            printf("#################################################################\n");
            printf("    Sample %d of %d, elapsed time %.2f sec. ETA: %.0f sec   \n",i,samples,ct,(samples-i)*ct/(double)i);
            printf("#################################################################\n");
            maxClusterCount=0;
            clusterCount=0;
            totBreaks=0;
            nwf_broken=false;
            
            runSimulationNoGraphics();
            // 
            //  The following code is processing the simulation data
            //
            manyPerimeters.clear();
            maxCC.push_back(maxClusterCount);
            double burstInit=straincurve[0];
            int burstSize=1;
            int k=1;
            while(k<N){
                while(straincurve[k]<burstInit){
                    if(k == N-1) {
                        fatalBurst[N-burstSize]++;
                        break;}

                    burstSize+=1;
                    k+=1;
                }
                burst[burstSize]++;
                burstInit=straincurve[k];
                burstSize=1;
                k+=1;
            }
            oldMeanStraincurve=meanStraincurve;
            for (int j = 0; j < N; ++j)
                meanStraincurve[j]+=(straincurve[j]-meanStraincurve[j])/((double) i);
            for (int j = 0; j < N; ++j)
                stdstrain[j]+=(straincurve[j]-meanStraincurve[j])*(straincurve[j]-oldMeanStraincurve[j]);
            for (int j = 0; j < N; ++j)
                meanFiberElongation[j]+=(fiberElongation[j]-meanFiberElongation[j])/((double) i);
            hist[nextWeakestFiber]++;
            if ( (ct - last_saving_time) > 3600) //saves about every hour
            {
                //save_to_file(i,meanStraincurve,stdstrain,burst,maxCC, hist,fatalBurst,meanFiberElongation);
                last_saving_time=ct;
            }
        }
        clock_t tot_end=clock();
        double tot_secs=(tot_end-tot_begin)/(double) CLOCKS_PER_SEC;
        printf("#################################################################\n");
        printf("                    Finished, total time %f sec \n",tot_secs);
        printf("#################################################################\n");
        //save_to_file(samples,meanStraincurve,stdstrain,burst,maxCC,hist,fatalBurst,meanFiberElongation); 
    #endif
    return 0;
}

#ifndef GRAPHICS
void runSimulationNoGraphics(){
    printf("System size %d t %d (%d) \n", Lx,Ly, N);
    fibers=new Fiber[N];
    int xx;
    int yy;
    noperimeter.resize(N); //allocate noperimeter
    clock_t begin = clock();
    for (int i = 0; i < N; i++) {
        xx = i % Lx;
        yy = i / Lx;
        fibers[i].t = 1.0/( cutoff - log(rng()) );
        fibers[i].fiberID=i;
        noperimeter[i]=&fibers[i];

        //Initialize the fibers
        for (int j = 0; j < 4; ++j)
        {
            fibers[i].height[j]=1; //Default height is 1, multiple conventions exists.
            fibers[i].left[j]=NULL;
            fibers[i].right[j]=NULL;
            fibers[i].myPerimeter[j]=NULL;
        }
    }
    clock_t end = clock();
    double elapsed_secs = double(end - begin) / CLOCKS_PER_SEC;

    std::sort(noperimeter.begin(), noperimeter.end(), lowestThreshold); //Sort fibers in noperimeter wrt. the true theshold 
    printf("Strongest sigma/x %f, weakest sigma/x is %f \n",noperimeter.front()->t,noperimeter.back()->t);
    printf("******************************\n");
    printf("Running simulation\n\n");
    begin = clock();
    clock_t current;
    double secs;
    nwf_broken = false; //Used to keep track of the time step at which the next weakest fiber break
    for(int i=0;i<(N);i++){
        if(totBreaks%(100)==0) {
            current=clock();
            secs= double(current - begin) / CLOCKS_PER_SEC; 
            printf("\rTime:  %f, progress: %f %%", secs, 100*totBreaks/(double)N);
        }
        if(totBreaks==N) break; 
        current=clock();
        secs= double(current - tot_time) / CLOCKS_PER_SEC; 
        breakWeakest();
        }
    end = clock();
    elapsed_secs = double(end - begin) / CLOCKS_PER_SEC;
    printf("\r******************************                           \n");
    printf("    Elapsed time: %f sec\n", elapsed_secs);
    printf("******************************\n");
    delete [] fibers;
}
#endif

void breakWeakest(){ //As all updates are done subsequent to a fiber break, it is easy to identify the next fiber to break
    //Three alternatives must be checked:
    //  #1 noPerLoad: the load on fibers that are not neighbouring any broken fibers
    //  #2 perLoad: the load on fibers included in at least one perimeter
    //  #3 manyPerLoad: the load on fibers included in at least one additional perimeter
    double noPerLoad = 0;
    while(not noperimeter.empty()){
        if(noperimeter.back()->status==BROKEN){
            noperimeter.pop_back();
            if(!nwf_broken){
                nextWeakestFiber=totBreaks;
                nwf_broken=true;
            }
        }else{
            noPerLoad=noperimeter.back()->t;   
            break;
        }
    }   
    Fiber *sp=NULL;//weakest fiber in single perimeter (one or more )
    double perLoad = 0;
    if(rootPerimeter!=NULL){
        sp=findmax(rootPerimeter)->weakest();
        perLoad=sp->trueload();
    }
    if(totBreaks<int(N*treeToVec)){
        weakestFiberManyPerLoad=0;
        weakestFiberManyPer=NULL;
        if(manyPerimeterRoot) {
            ManyPerimeter *mp=findmax(manyPerimeterRoot);
            weakestFiberManyPer=mp->f;
            weakestFiberManyPerLoad=mp->trueload;
        }
    }else{
        updateManyPerVector();
    }
    if(weakestFiberManyPer!=NULL && weakestFiberManyPerLoad > perLoad){
        straincurve[totBreaks]=1/(double)weakestFiberManyPerLoad;
        fiberElongation[totBreaks]=1/weakestFiberManyPer->t;
        weakestFiberManyPer->Break();
    } else if (perLoad>noPerLoad){
        straincurve[totBreaks]=1/perLoad;
        fiberElongation[totBreaks]=1/sp->t;//perLoad;
        sp->Break();
    }else { //if(noPerLoad>perLoad and noPerLoad>weakestFiberManyPerLoad)
        straincurve[totBreaks]=1/noPerLoad;
        fiberElongation[totBreaks]=1/noPerLoad;
        noperimeter.back()->Break();
        noperimeter.pop_back();
        if(totBreaks>1 && !nwf_broken){
            nextWeakestFiber=totBreaks;
            nwf_broken=true;
        }
    }
}
void break_ELS(){ //Equal Load Sharing
	noperimeter.back()->Break();
	noperimeter.pop_back();
}
void makeNewSeedFile(){ //creates the file "seed_binary.bin"
    uint64_t old_s[ 16 ];
    char old_p=p;

    for (int i = 0; i < 16; i++){
        s[i]=(unsigned int) rngi();
        old_s[i]=s[i];
    }
    for (int i = 0; i < 16; i++){
        rng();
    }
    FILE *fp = fopen("seed_binary.bin", "wb");
    for (int j = 0; j < 1; j++) {
        int buffer_size = 2+2+4+4+1+16*8;
        unsigned char buffer[buffer_size];
        *(short*)(buffer+0) = (short)Lx;
        *(short*)(buffer+2) = (short)Ly;
        *(float*)(buffer+4) = (float)cutoff;
        *(unsigned long int*)(buffer+8) = (unsigned long int)samples;
        *(unsigned char*)(buffer+12) = (unsigned char)p;
        for (int i = 0; i < 16; i++) {
            *(uint64_t*)(buffer+13+8*i) = (uint64_t)s[i];
        }
        fwrite(buffer, sizeof(unsigned char), buffer_size, fp);
        for (int i = 0; i < Lx*Ly*samples; i++) {
            rng();
        }
    }
    fclose(fp);
}
void initializeSeed(){
    int n;
    uint64_t old_s[ 16 ];
    char old_p;
    //First chech if the file exists
    if ( access("seed_binary.bin", F_OK ) == -1 ) {
        makeNewSeedFile();
    } 
    FILE *fp=fopen("seed_binary.bin", "rb");
    while (!feof(fp)) { 
        int buffer_size = 2+2+4+4+1+16*8;
        unsigned char oldbuffer[buffer_size];
        n=fread(oldbuffer, sizeof(unsigned char), buffer_size, fp);
        if(n==0) break;
        p=(int)(*(unsigned char*)(oldbuffer+12));
        for (int i = 0; i < 16; i++) {
          s[i]=(uint64_t)(*(uint64_t*)(oldbuffer+13+8*i));
        } 
    }
    fclose(fp);
    old_p=p;
    for (int i = 0; i < 16; i++){
        old_s[i]=s[i];
    }
    for (int i = 0; i < Lx*Ly*samples; i++){
        next();
    }
    fp = fopen("seed_binary.bin", "ab");
    int buffer_size = 2+2+4+4+1+16*8;
    unsigned char buffer[buffer_size];
    *(short*)(buffer+0) = (short)Lx;
    *(short*)(buffer+2) = (short)Ly;
    *(float*)(buffer+4) = (float)cutoff;
    *(unsigned long int*)(buffer+8) = (unsigned long int)samples;
    *(unsigned char*)(buffer+12) = (unsigned char)p;
    for (int i = 0; i < 16; i++) {
        *(uint64_t*)(buffer+13+8*i) = (uint64_t)s[i];
    }
    fwrite(buffer, sizeof(unsigned char), buffer_size, fp);
    fclose(fp);
    p=old_p;
    for (int i = 0; i < 16; i++){
        s[i]=old_s[i];
    }   
}
void save_to_file(int sample,std::vector<double> &meanStraincurve,std::vector<double> &stdstrain,std::map<int, int> &burst,std::vector<int> &new_cc,std::map<int, int> &hist,std::map<int, int> &fatalBurst,std::vector<double> &meanFiberElongation){
        remove(strainfilename);
        snprintf (strainfilename, sizeof strainfilename, "straincurve_%05d_%.5f_%07d.dat", Lx,cutoff,sample);
        FILE *sc=fopen(strainfilename,"w");

        for (int i = 0; i < N; ++i)
        {
            if (sample>1)            fprintf(sc, "%.10f %f \n", meanStraincurve[i],sqrt(stdstrain[i]/(sample-1)));
            else fprintf(sc, "%.10f %f \n", meanStraincurve[i],sqrt(stdstrain[i]));
        }
        fclose(sc);
        //read old (max) cluster count old_cc
        std::vector<int> old_cc;
        int tmp;
        FILE *ccFile = fopen(ccfilename, "r+");
        if(ccFile){
            for ( ;; ){
                fscanf(ccFile, "%d", &tmp);
                if( feof(ccFile) ) break;
                old_cc.push_back(tmp);;
            }
            fclose(ccFile);
            remove(ccfilename);
        }
        snprintf (ccfilename, sizeof ccfilename, "maxclustercount_%05d_%.5f_%07d.dat", Lx,cutoff,sample);
        ccFile = fopen(ccfilename, "w");
        for (int i = 0; i < old_cc.size(); ++i)
        {
            fprintf(ccFile,"%d \n", old_cc[i]);
        }
        for (int i = 0; i < new_cc.size(); ++i)
        {
            fprintf(ccFile,"%d \n", new_cc[i]);
        }
        fclose(ccFile);
        old_cc.clear();
        remove(nwffilename);
        snprintf (nwffilename, sizeof nwffilename, "nwf_%05d_%.5f_%07d.dat", Lx,cutoff,sample);
        FILE *nwfFile = fopen(nwffilename, "w");
        for (auto it=hist.begin(); it!=hist.end(); ++it) {
            fprintf(nwfFile,"%d %d \n", it->first,it->second);
        }
        fclose(nwfFile);
        remove(burstfilename);
        snprintf (burstfilename, sizeof burstfilename, "burst_%05d_%.5f_%07d.dat", Lx,cutoff,sample);
        FILE *burstFile = fopen(burstfilename, "w");
        double bstd;
        for (auto it=burst.begin(); it!=burst.end(); ++it) {
            fprintf(burstFile,"%d %d \n", it->first,it->second);
        }
        fclose(burstFile);
        remove(fburstfilename);
        snprintf (fburstfilename, sizeof fburstfilename, "fatalburst_%05d_%.5f_%07d.dat", Lx,cutoff,sample);
        FILE *fBfile = fopen(fburstfilename, "w");
        for (auto it=fatalBurst.begin(); it!=fatalBurst.end(); ++it) {
            fprintf(fBfile,"%d %d \n", it->first,it->second);
        }
        fclose(fBfile);
        remove(stressfilename);
        snprintf (stressfilename, sizeof stressfilename, "fiberelongation_%05d_%.5f_%07d.dat", Lx,cutoff,sample);
        FILE *stressFile=fopen(stressfilename,"w");
        for (int i = 0; i < N; ++i)
        {
            fprintf(stressFile, "%.10f \n", meanFiberElongation[i]);
        }
        fclose(stressFile);
}